<?php
class cmsCache {

    private static $instance;

    private $cacher;

    public $query_count = 0;


    public static function getInstance() {
        if (self::$instance === null) {
            self::$instance = new self;
        }
        return self::$instance;
    }

    public function __construct() {

        $cacher_class = 'cmsCache' . string_to_camel('_', cmsConfig::get('cache_method'));

        $this->cacher = new $cacher_class();

    }

    public function start() {
        if (!cmsConfig::get('cache_enabled')) { return false; }
        $this->cacher->start();
    }

    public function stop() {
        if (!cmsConfig::get('cache_enabled')) { return false; }
        $this->cacher->stop();
    }

    public function set($key, $value, $ttl=false){

        $config = cmsConfig::getInstance();
        if (!$config->cache_enabled) { return false; }
        if (DEBUG_ON) {
          $debug = cmsDebug::getInstance();
          $debug->beginCache();
        }
        if (!$ttl) { $ttl = $config->cache_ttl; }
        $ret = $this->cacher->set($key, $value, $ttl);
        if (DEBUG_ON) $debug->endCache("set", $key, $ret);
        return $ret;

    }

    public function has($key){

        if (!cmsConfig::get('cache_enabled')) { return false; }

        return $this->cacher->has($key);

    }

    public function get($key){

        if (!cmsConfig::get('cache_enabled')) { return false; }
        if (DEBUG_ON) {
          $debug = cmsDebug::getInstance();
          $debug->beginCache();
        }
        if (!$this->has($key)) {
          if (DEBUG_ON) $debug->endCache("get", $key, "Not in cache");
          return false;
        }
        $value = $this->cacher->get($key);
        if (cmsConfig::get('debug') && $value) {
            $this->query_count++;
        }
        if (DEBUG_ON) $debug->endCache("get", $key, (is_array($value) ? count($value, COUNT_RECURSIVE) : 0));
        return $value;

    }

    public function clean($key){

        if (!cmsConfig::get('cache_enabled')) { return false; }

        return $this->cacher->clean($key);

    }

}
